package flink.zanxiangnet.ad.monitoring.maxcompute.sink;

import com.aliyun.odps.PartitionSpec;
import com.aliyun.odps.tunnel.TableTunnel;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeWriter;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.CollectionUtils;

import java.util.*;
import java.util.concurrent.*;

/**
 * 缓存要写入 MaxCompute的数据
 */
public class MaxComputeSinkBuffer<T> {
    private static final Logger log = LoggerFactory.getLogger(MaxComputeSinkBuffer.class);

    private final TableTunnel tunnel;
    private final String projectName;
    private final String tableName;
    private final long buffRefreshTime;
    private final long maxBuffer;
    private final ExecutorService threadPool;
    private final Map<String, List<T>> partitionData;
    private final Class<T> clazz;

    private Thread coreThread;

    public MaxComputeSinkBuffer(
            long buffRefreshTime,
            long maxBuffer,
            int partitionCount,
            Class<T> clazz,
            TableTunnel tunnel,
            String projectName,
            String tableName
    ) {
        this.clazz = clazz;
        this.tunnel = tunnel;
        this.projectName = projectName;
        this.tableName = tableName;
        this.buffRefreshTime = buffRefreshTime;
        this.maxBuffer = maxBuffer;
        partitionData = new ConcurrentHashMap<>(partitionCount);

        threadPool = new ThreadPoolExecutor(
                partitionCount,
                partitionCount,
                0L,
                TimeUnit.MILLISECONDS,
                new LinkedBlockingQueue<>(),
                new ThreadFactoryBuilder()
                        .setNameFormat("maxcompute-writer-%d").build());
        start();
    }

    /**
     * 写入数据
     * @param partition 数据分区
     * @param t 数据
     */
    public void put(String partition, T t) {
        List<T> list = partitionData.computeIfAbsent(partition, key -> new ArrayList<>());
        list.add(t);
    }

    public void close() {
        coreThread.interrupt();
    }

    private void start() {
        coreThread = new Thread(() -> {
            Map<String, Long> lastSaveTime = new HashMap<>();
            // 开启一个定时轮询缓存数据的线程
            while (true) {
                try {
                    if (CollectionUtils.isEmpty(partitionData)) {
                        // 没有数据，等 5s
                        Thread.sleep(5000L);
                        continue;
                    }
                    long now = System.currentTimeMillis();
                    List<String> saveKeys = new ArrayList<>(partitionData.size());
                    for (Map.Entry<String, List<T>> entry : partitionData.entrySet()) {
                        Long lastSaveTimeOfKey = lastSaveTime.computeIfAbsent(entry.getKey(), key -> now);
                        boolean save = now - lastSaveTimeOfKey > buffRefreshTime;
                        if (save || entry.getValue().size() > maxBuffer) {
                            saveKeys.add(entry.getKey());
                        }
                    }
                    if (CollectionUtils.isEmpty(saveKeys)) {
                        Thread.sleep(5000L);
                        continue;
                    } else {
                        System.out.println("===>实际写入数据2:" + JsonUtil.toString(saveKeys));
                        for (int i = 0; i < saveKeys.size(); i++) {
                            String key = saveKeys.get(i);
                            List<T> datas = partitionData.remove(key);
                            System.out.println("===>实际写入数据3:" + datas.size());
                            lastSaveTime.remove(key);
                            TableTunnel.UploadSession uploadSession = tunnel.createUploadSession(projectName, tableName, new PartitionSpec(key));
                            threadPool.submit(new MaxComputeWriter<>(clazz, uploadSession, uploadSession.openRecordWriter(i), datas));
                        }
                    }
                } catch (Exception e) {
                    log.error(e.getMessage(), e);
                }
            }
        });
        coreThread.start();
    }
}
