package flink.zanxiangnet.ad.monitoring.util.bean;

import org.apache.commons.lang3.StringUtils;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class BeanUtil {
    private static final Pattern humpPattern = Pattern.compile("[A-Z]");
    private static final Pattern linePattern = Pattern.compile("_(\\w)");

    /**
     * 解析 javabean的字段
     * @param clazz
     * @return
     */
    public static List<FieldInfo> parseBeanField(Class<?> clazz) {
        List<Field> fieldList = new ArrayList<>();
        Class<?> tempClazz = clazz;
        while (tempClazz != null) {
            Field[] fields = tempClazz.getDeclaredFields();
            if (fields.length != 0) {
                fieldList.addAll(Arrays.asList(fields));
            }
            tempClazz = tempClazz.getSuperclass();
        }
        if (fieldList.size() == 0) {
            throw new RuntimeException("Failed parsed class[" + clazz.getName() + "]. No Field!");
        }
        List<FieldInfo> fieldInfoList = new ArrayList<>(fieldList.size());
        for(Field field : fieldList) {
            DataMapping annotation = field.getAnnotation(DataMapping.class);
            FieldInfo fieldInfo = new FieldInfo();
            fieldInfo.setField(field);
            fieldInfo.setFieldName(field.getName());
            try {
                fieldInfo.setGetMethod(methodOfGet(clazz, field));
                fieldInfo.setSetMethod(methodOfSet(clazz, field));
            } catch (NoSuchMethodException e) {
                throw new RuntimeException("Failed parsed class[" + clazz.getName() + "]. The get or set method of [" + field.getName() + "] was not found!");
            }
            if (annotation != null) {
                if (annotation.ignore()) {
                    continue;
                }
                if (StringUtils.isNotBlank(annotation.columnName())) {
                    fieldInfo.setColumnName(annotation.columnName());
                } else {
                    fieldInfo.setColumnName(humpToLine(fieldInfo.getFieldName()));
                }
                fieldInfo.setColumnType(annotation.dataType());
            } else {
                fieldInfo.setColumnName(humpToLine(fieldInfo.getFieldName()));
                DataTypeEnum columnType = null;
                if (fieldInfo.getFieldType() == Byte.class) {
                    columnType = DataTypeEnum.INT;
                } else if (fieldInfo.getFieldType() == Integer.class) {
                    columnType = DataTypeEnum.INT;
                } else if (fieldInfo.getFieldType() == Long.class) {
                    columnType = DataTypeEnum.BIGINT;
                } else if (fieldInfo.getFieldType() == Float.class) {
                    columnType = DataTypeEnum.DOUBLE;
                } else if (fieldInfo.getFieldType() == Double.class) {
                    columnType = DataTypeEnum.DOUBLE;
                } else if (fieldInfo.getFieldType() == BigDecimal.class) {
                    columnType = DataTypeEnum.DECIMAL;
                } else if (fieldInfo.getFieldType() == Date.class) {
                    columnType = DataTypeEnum.DATETIME;
                } else if (fieldInfo.getFieldType() == LocalDate.class) {
                    columnType = DataTypeEnum.DATE;
                } else if (fieldInfo.getFieldType() == LocalDateTime.class) {
                    columnType = DataTypeEnum.DATETIME;
                } else if (fieldInfo.getFieldType() == Boolean.class) {
                    columnType = DataTypeEnum.BOOLEAN;
                } else if (fieldInfo.getFieldType() == String.class) {
                    columnType = DataTypeEnum.STRING;
                } else {
                    throw new RuntimeException("Failed parse class[" + clazz.getName() + "]. The field [" + field.getName() + "] type unknown!");
                }
                fieldInfo.setColumnType(columnType);
            }
            fieldInfoList.add(fieldInfo);
        }
        return fieldInfoList;
    }

    /**
     * 驼峰转下划线
     * @param str
     * @return
     */
    public static String humpToLine(String str) {
        Matcher matcher = humpPattern.matcher(str);
        StringBuffer sb = new StringBuffer();
        while (matcher.find()) {
            matcher.appendReplacement(sb, "_" + matcher.group(0).toLowerCase());
        }
        matcher.appendTail(sb);
        return sb.toString();
    }

    /**
     * 下划线转驼峰
     * @param str
     * @return
     */
    public static String lineToHump(String str) {
        str = str.toLowerCase();
        Matcher matcher = linePattern.matcher(str);
        StringBuffer sb = new StringBuffer();
        while (matcher.find()) {
            matcher.appendReplacement(sb, matcher.group(1).toUpperCase());
        }
        matcher.appendTail(sb);
        return sb.toString();
    }

    public static <T> Method methodOfGet(Class<T> clazz, Field field) throws NoSuchMethodException {
        String fieldName = field.getName();
        String methodName = "get" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
        return clazz.getMethod(methodName);
    }

    public static <T> Method methodOfSet(Class<T> clazz, Field field) throws NoSuchMethodException {
        String fieldName = field.getName();
        String methodName = "set" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
        return clazz.getMethod(methodName, field.getType());
    }
}
